use std::sync::{Arc, RwLock};

use iced::{
    Alignment::Center,
    Element, Task, Theme,
    widget::{button, column, row},
};
use protocol::Error;
use tokio::sync::{
    broadcast::{self},
    oneshot,
};

use crate::{
    ClientConfig, MICROPHONE_BUFFER_LENGHT,
    stream::{connect, disconnect_watcher},
    voice::record,
};

#[derive(Debug, Default)]
struct Signal {
    microphone_stop_sender: Option<oneshot::Sender<bool>>,
    // speaker: Option<oneshot::Sender<bool>>,
    connection_stop_sender: Option<oneshot::Sender<bool>>,
}

impl Signal {
    fn reset_microphone(&mut self) -> Result<(), Error> {
        if let Some(microphone_signal) = &self.microphone_stop_sender {
            if !microphone_signal.is_closed() {
                self.microphone_stop_sender
                    .take()
                    .expect("Never")
                    .send(true)
                    .unwrap();
                self.microphone_stop_sender = None;
                return Ok(());
            }
        }
        Err(Error::Signal("Reset".to_string()))
    }

    fn reset_connection(&mut self) -> Result<(), Error> {
        if let Some(connection_signal) = &self.connection_stop_sender {
            if !connection_signal.is_closed() {
                self.connection_stop_sender
                    .take()
                    .expect("Never")
                    .send(true)
                    .unwrap();
                self.connection_stop_sender = None;
                return Ok(());
            }
        }
        Err(Error::Signal("Reset".to_string()))
    }
}

#[derive(Debug)]
struct Channel {
    microphone: Arc<broadcast::Sender<f32>>,
    // speaker: (broadcast::Sender<f32>, broadcast::Receiver<f32>),
}

impl Channel {
    fn new() -> Self {
        Self {
            microphone: broadcast::channel(MICROPHONE_BUFFER_LENGHT / 4).0.into(),
            // speaker: broadcast::channel(BUFFER_LENGTH),
        }
    }
}

#[derive(Debug, Default, Clone, Copy)]
struct GUIStatus {
    room: State,
    microphone: State,
}

#[derive(Debug, Clone, Copy)]
pub enum State {
    Active,
    Passive,
    Loading,
}
impl Default for State {
    fn default() -> Self {
        State::Passive
    }
}

#[derive(Debug, Clone, Copy)]
pub enum Message {
    None,
    JoinRoom,
    LeaveRoom,
    UnmuteMicrophone,
    MuteMicrophone,
}

#[derive(Debug)]
pub struct App {
    client_config: Arc<ClientConfig>,
    gui_status: Arc<RwLock<GUIStatus>>,
    channel: Channel,
    signal: Signal,
}

impl App {
    fn reset_microphone(&mut self) -> Result<(), Error> {
        self.signal.reset_microphone()?;
        self.gui_status.write().unwrap().microphone = State::Passive;
        Ok(())
    }

    fn reset_connection(&mut self) -> Result<(), Error> {
        self.signal.reset_connection()?;
        self.gui_status.write().unwrap().room = State::Passive;
        Ok(())
    }

    pub fn theme(&self) -> Theme {
        Theme::Dark
    }

    pub fn new() -> Self {
        App {
            client_config: ClientConfig::new().into(),
            gui_status: RwLock::new(GUIStatus::default()).into(),
            channel: Channel::new(),
            signal: Signal::default(),
        }
    }
    pub fn view(&self) -> Element<'_, Message> {
        let join_room_button = match self.gui_status.read().unwrap().room {
            State::Active => button("Leave Room").on_press(Message::LeaveRoom),
            State::Passive => button("Join Room").on_press(Message::JoinRoom),
            State::Loading => button("Loading"),
        };

        let microphone_button = match self.gui_status.read().unwrap().microphone {
            State::Active => button("Mute").on_press(Message::MuteMicrophone),
            State::Passive => button("Unmute").on_press(Message::UnmuteMicrophone),
            State::Loading => button("Loading"),
        };
        column![
            row![join_room_button, microphone_button]
                .spacing(20)
                .align_y(Center)
        ]
        .align_x(Center)
        .into()
    }
    pub fn update(&mut self, message: Message) -> Task<Message> {
        match message {
            Message::None => Task::none(),
            Message::JoinRoom => {
                self.gui_status.write().unwrap().room = State::Loading;
                let client_config = self.client_config.clone();
                let gui_status = self.gui_status.clone();
                let microphone_receiver = self.channel.microphone.subscribe();

                let (connection_stop_sender, connection_stop_receiver) = oneshot::channel();
                self.signal.connection_stop_sender = Some(connection_stop_sender);
                Task::perform(
                    async move {
                        match connect(microphone_receiver, client_config).await {
                            Ok(connection_return) => {
                                tokio::spawn(disconnect_watcher(
                                    connection_stop_receiver,
                                    connection_return,
                                ));
                                gui_status.write().unwrap().room = State::Active;
                            }
                            Err(err_val) => {
                                eprintln!("Error: Connect | {}", err_val);
                                gui_status.write().unwrap().room = State::Passive;
                            }
                        }
                    },
                    |_| {},
                )
                .map(|_| Message::None)
            }
            Message::LeaveRoom => {
                self.gui_status.write().unwrap().room = State::Loading;
                if let Err(err_val) = self.reset_connection() {
                    eprintln!("Error: Leave Room | {}", err_val);
                }
                Task::none()
            }
            Message::UnmuteMicrophone => {
                self.gui_status.write().unwrap().microphone = State::Loading;
                let microphone_sender = self.channel.microphone.clone();
                let microphone_stop_signal = oneshot::channel();
                self.signal.microphone_stop_sender = Some(microphone_stop_signal.0);
                let is_microphone_started_signal = oneshot::channel();
                tokio::spawn(record(
                    microphone_sender,
                    is_microphone_started_signal.0,
                    microphone_stop_signal.1,
                ));
                let gui_status = self.gui_status.clone();
                Task::perform(
                    async move {
                        if let Ok(_) = is_microphone_started_signal.1.await {
                            gui_status.write().unwrap().microphone = State::Active;
                        }
                    },
                    |_| Message::None,
                )
            }
            Message::MuteMicrophone => {
                self.gui_status.write().unwrap().microphone = State::Loading;
                if let Err(err_val) = self.reset_microphone() {
                    eprintln!("Error: Mute Microphone | {}", err_val);
                }
                Task::none()
            }
        }
    }
}
